"""
Copyright (c) 2021, Zolisa Bleki
SPDX-License-Identifier: BSD-3-Clause
"""
from collections import defaultdict

from imblearn.over_sampling.base import BaseOverSampler
from imblearn.utils import Substitution
from imblearn.utils._docstring import (
    _random_state_docstring,
    _n_jobs_docstring,
)
from imblearn.utils._validation import check_neighbors_object
import numpy as np
from sklearn.base import clone
from sklearn.mixture import GaussianMixture

from ._common import check_random_state, safe_random_state


@Substitution(
    sampling_strategy=BaseOverSampler._sampling_strategy_docstring,
    random_state=_random_state_docstring,
    n_jobs=_n_jobs_docstring
)
class GAMUS(BaseOverSampler):
    """
    Gaussian Mixture Model based Adaptive Minority up-sampling.

    Parameters
    ----------
    {sampling_strategy}
    n_neighbors : int or estimator object, default=5
        If ``int``, number of nearest neighbours within the minority class per
        minority class data point. If object, an estimator that inherits from
        :class:`~sklearn.neighbors.base.KNeighborsMixin` that will be used to
        find the neighbors.
    n_pickedup : int, default=20
        The number of synthetic samples generated for each minority sample and
        neighbor pair. This is equivalent to ``r`` parameter in [1]_.
    n_clusters : int or estimator, default=2
        If ``int``, The number of components used to fit the Gaussian Mixture
        Model on the minority class synthetic data points. If object, an
        estimator from the ``sklearn.mixture`` module that will be used to fit
        the Gaussian Mixture Model.
    gmm_params : dict, default=None
        Non-default parameters to pass to the :class:`~sklearn.mixture.GaussianMixture`
        instance used to fit the Gaussian Mixture Model to the data, if
        `n_cluster` is an integer. This parameter is ignored if `n_cluster` is
        an estimator object.
    proba_threshold : float, default=0.5
        A probability threshold associated with the posterior probability of a
        majority class sample belonging to the i'th Gaussian Mixture cluster.
        If the probability is above the threshold then a majority data point
        will be assigned the cluster associated with the probability.
    weight_threshold : float, default=0.5
        A threshold associated with the proportion of majority samples that are
        assigned to a particular cluster using `proba_threshold`. If the
        proportion of majority samples assigned to a particular cluster is
        above `weight_threshold` then the weight of that cluster is assigned
        `weight_threshold`, else 0.
    {random_state}
    {n_jobs}

    References
    ----------
    .. [1] A. Tripathi, R. Chakraborty and S. K. Kopparapu, "A Novel Adaptive
           Minority Oversampling Technique for Improved Classification in Data
           Imbalanced Scenarios," 2020 25th International Conference on Pattern
           Recognition (ICPR), 2021, pp. 10650-10657,
           doi: 10.1109/ICPR48806.2021.9413002.

    Examples
    --------
    >>> from pyloras import GAMUS
    >>> from sklearn.datasets import make_classification
    >>> from collections import Counter
    >>> l = GAMUS()
    >>> X, y = make_classification(n_classes=3, class_sep=3,
    ... weights=[0.7, 0.2, 0.1], n_informative=3, n_redundant=1, flip_y=0,
    ... n_features=20, n_clusters_per_class=1, n_samples=2000, random_state=10)
    >>> print('Original dataset shape %s' % Counter(y))
    Original dataset shape Counter({{1: 400, 2: 200, 0: 1400}})
    >>> X_res, y_res = l.fit_resample(X, y)
    >>> print(f"Resampled dataset shape % Counter(y_res))
    Resampled dataset shape Counter({{1: 1400, 2: 1400, 0: 1400}})

    """
    def __init__(
        self,
        *,
        sampling_strategy="auto",
        n_neighbors=5,
        n_pickedup=10,
        n_clusters=2,
        gmm_params=None,
        proba_threshold=0.5,
        weight_threshold=0.5,
        random_state=None,
        n_jobs=None,
    ):
        super().__init__(sampling_strategy=sampling_strategy)
        self.n_neighbors = n_neighbors
        self.n_pickedup = n_pickedup
        self.n_clusters = n_clusters
        self.gmm_params = gmm_params
        self.proba_threshold = proba_threshold
        self.weight_threshold = weight_threshold
        self.random_state = random_state
        self.n_jobs = n_jobs

    def _validate_estimator(self, rng):
        if not 0 < self.proba_threshold < 1:
            raise ValueError("proba_threshold must be between 0 and 1")
        if not 0 < self.weight_threshold < 1:
            raise ValueError("weight_threshold must be between 0 and 1")

        self.nn_ = check_neighbors_object(
            "n_neighbors", self.n_neighbors, additional_neighbor=1
        )
        self.nn_.set_params(n_jobs=self.n_jobs, algorithm="ball_tree")
        self.alpha_ = np.arange(1, self.n_pickedup + 1) / self.n_pickedup

        if not isinstance(self.n_clusters, int):
            gmm_methods = {'fit', 'predict_proba', 'set_params'}
            for attr in gmm_methods:
                if not hasattr(self.n_clusters, attr):
                    raise ValueError(
                        f'The GMM estimator must implement all of {gmm_methods}'
                    )
            self.gmm_ = clone(self.n_clusters)
            self.gmm_.set_params(random_state=safe_random_state(rng))
        else:
            self.gmm_ = GaussianMixture(
                n_components=self.n_clusters,
                random_state=safe_random_state(rng),
            )
            if self.gmm_params:
                self.gmm_.set_params(**self.gmm_params)

    def _fit_resample(self, X, y):
        random_state = check_random_state(self.random_state)
        self._validate_estimator(random_state)
        X_res = [X.copy()]
        y_res = [y.copy()]

        gamus_samples = defaultdict(lambda: [])
        n_features = X.shape[1]

        for minority_class, samples_to_make in self.sampling_strategy_.items():
            if samples_to_make == 0:
                continue

            mask = (y == minority_class)
            Xmin = X[mask]
            self.nn_.fit(Xmin)
            nns = self.nn_.kneighbors(Xmin, return_distance=False)[:, 1:]

            XX = Xmin[:, None]
            # shape: (n_pickedup, Xmin.shape[0], n_neighbors, n_features)
            scaled_diffs = np.multiply.outer(self.alpha_, XX - Xmin[nns])
            synth_minority = (XX + scaled_diffs).reshape(-1, n_features)

            self.gmm_.fit(np.concatenate((Xmin, synth_minority)))

            Xmaj = X[~mask]
            q = (self.gmm_.predict_proba(Xmaj) > self.proba_threshold).sum(axis=0)
            weights = 1.0 - q / Xmaj.shape[0]
            # ensure if all majority samples are assigned to one cluster then
            # the weight is set to 1
            weights[q == Xmaj.shape[0]] = 1.
            weights[np.argwhere(weights < self.weight_threshold)] = 0
            self.cluster_weights_ = weights
            samples_per_cluster = np.ceil(
                samples_to_make * weights / sum(weights)
            ).astype(int)
            minority_cluster_labels = (
                np.argmax(self.gmm_.predict_proba(synth_minority), axis=1)
            )
            # TODO: Vectorize this loop if possible
            for i, c in enumerate(samples_per_cluster):
                mat = synth_minority[minority_cluster_labels == i]
                if not c or not mat.shape[0]:
                    continue
                replace = mat.shape[0] < c
                gamus_samples[minority_class].append(
                    random_state.choice(mat, size=c, replace=replace, axis=0)
                )

            samples_to_drop = sum(samples_per_cluster) - samples_to_make
            random_state.shuffle(gamus_samples[minority_class])
            X_res.append(
                np.vstack(gamus_samples[minority_class])[samples_to_drop:]
            )
            y_res.append([minority_class] * X_res[-1].shape[0])

        return np.concatenate(X_res), np.concatenate(y_res)



"""
Copyright (c) 2021, Zolisa Bleki
SPDX-License-Identifier: BSD-3-Clause
"""
from collections import defaultdict
import math
from numbers import Number
from typing import NamedTuple

from imblearn.over_sampling.base import BaseOverSampler
from imblearn.utils import Substitution
from imblearn.utils._docstring import (
    _random_state_docstring,
    _n_jobs_docstring,
)
from imblearn.utils._validation import check_neighbors_object
import numpy as np

from ._common import check_random_state


class MinorityData(NamedTuple):
    # Minority class data points
    X: np.ndarray
    # Non-overlapping partitions of the minority class data points
    clusters: list
    # Number of prowras synthetic samples to generate per cluster
    n_new_samples: list
    # Number of convex combinations of each new synthetic sample per cluster
    n_affine: list


@Substitution(
    sampling_strategy=BaseOverSampler._sampling_strategy_docstring,
    random_state=_random_state_docstring,
    n_jobs=_n_jobs_docstring
)
class ProWRAS(BaseOverSampler):
    """Proximity Weighted Random Affine Shadowsampling (ProWRAS).

    Parameters
    ----------
    {sampling_strategy}
    max_clusters : int, default=5
        The maximum number of clusters to partition the data points of each
        class included in ``sampling_strategy``. This parameter is equivalent
        to ``max_levels`` in [1]_.
    n_neighbors_max : int or estimator object, default=None
        If ``int``, the maximum number of minority class nearest neighbours of
        all the majority class data points. If object, an estimator that
        inherits from :class:`~sklearn.neighbors.base.KNeighborsMixin`
        that will be used to find the neighbors. If None, a maximum of 5
        neighbors is used.
    decay_rate : int, default=1
        The rate of decay of the weights assigned to the clusters formed by the
        data points of the minority class(es). Defaults to 1.
    n_cluster_neighbors : int or estimator object, default=5
        If ``int``, the maximum size of the minority class neighborhood per
        cluster that is used to generate the synthetic samples. If object, an
        estimator that inherits from :class:`~sklearn.neighbors.base.KNeighborsMixin`
        that will be used to find the neighbors. If larger than the size of the
        cluster itself, the whole cluster will be used as a neighbourhood.
        Defaults to 5. This parameter is equivalent to ``neb_conv`` in [1]_.
    max_affine : int, default=10
        The maximum number of shadow samples to use when generating the synthetic
        samples through random affine combinations. If ``2``, SMOTE-like
        sythetic samples will be generated that have high variance. If greater
        than 2, then LoRAS-like synthetic samples will be generated by taking
        a convex combination of them. In this case, the actual number of convex
        combinations for each cluster is decided by the normalized proximity
        weight of the respective cluster; which are then multiplied with the
        parameter to obtain the number of appropriate convex combinations of
        shadowsamples for each cluster.
    n_shadow : int, default=100
        The number of shadow samples to generate per minority class data point
        in the case where `max_affine` is greater than 2.
    std : float or sequence, default=0.0001
        The standard deviation of the Normal distribution to add to each data
        point feature when generating shadow samples. If the input is a
        sequence, its size must be equal to the number of features of ``X``
        when calling the ``fit_resample`` method. If ``float``, then same
        standard deviation will be used for all shadow samples generated.
    {random_state}
    {n_jobs}

    References
    ----------
    .. [1] Bej, S., Schultz, K., Srivastava, P., Wolfien, M., & Wolkenhauer,
       O. (2021). A multi-schematic classifier-independent oversampling
       approach for imbalanced datasets. ArXiv, abs/2107.07349.

    Examples
    --------
    >>> from pyloras import ProWRAS
    >>> from sklearn.datasets import make_classification
    >>> from collections import Counter
    >>> l = ProWRAS()
    >>> X, y = make_classification(n_classes=3, class_sep=3,
    ... weights=[0.7, 0.2, 0.1], n_informative=3, n_redundant=1, flip_y=0,
    ... n_features=20, n_clusters_per_class=1, n_samples=2000, random_state=10)
    >>> print('Original dataset shape %s' % Counter(y))
    Original dataset shape Counter({{1: 400, 2: 200, 0: 1400}})
    >>> X_res, y_res = l.fit_resample(X, y)
    >>> print(f"Resampled dataset shape % Counter(y_res))
    Resampled dataset shape Counter({{1: 1400, 2: 1400, 0: 1400}})

    """
    def __init__(
        self,
        *,
        sampling_strategy="auto",
        max_clusters=15,
        n_neighbors_max=None,
        decay_rate=1.,
        n_cluster_neighbors=5,
        max_affine=10,
        n_shadow=100,
        std=0.001,
        random_state=None,
        n_jobs=None,
    ):
        super().__init__(sampling_strategy=sampling_strategy)
        self.max_clusters = max_clusters
        self.n_neighbors_max = n_neighbors_max
        self.decay_rate = decay_rate
        self.n_cluster_neighbors = n_cluster_neighbors
        self.max_affine = max_affine
        self.n_shadow = n_shadow
        self.std = std
        self.random_state = random_state
        self.n_jobs = n_jobs

    def _make_cluster_data(self, minority_data, X_not_minority, samples_to_make):
        nn = check_neighbors_object(
            "n_neighbors_max", self.n_neighbors_max if self.n_neighbors_max else 5
        ).set_params(n_jobs=self.n_jobs)

        X = minority_data.X
        max_weight = 0
        weights = []
        for i in range(1, self.max_clusters):
            if X.shape[0] == 0:
                break
            k = min(X.shape[0], nn.n_neighbors)
            nn.set_params(n_neighbors=k).fit(X)
            cluster_indices = np.unique(
                nn.kneighbors(X_not_minority, return_distance=False)
            )
            minority_data.clusters.append(X[cluster_indices])
            weights.append(math.exp(-self.decay_rate * (i - 1)))
            max_weight = weights[-1] if weights[-1] > max_weight else max_weight
            X = X[~np.in1d(np.arange(X.shape[0]), cluster_indices)]

        # append the remaining partition (if it exists) as the last cluster
        # if the last (max_cluster - 1) clusters have been used up.
        if X.shape[0]:  # pragma: no cover
            weights.append(
                math.exp(-self.decay_rate * (self.max_clusters - 1))
            )
            max_weight = weights[-1] if weights[-1] > max_weight else max_weight
            minority_data.clusters.append(X)

        weight_sum = sum(weights)
        for i in range(len(weights)):
            # normalize weights
            weights[i] /= weight_sum
            minority_data.n_new_samples.append(
                math.ceil(samples_to_make * weights[i])
            )
            minority_data.n_affine.append(
                math.ceil(self.max_affine * weights[i] / max_weight)
            )
        return minority_data

    def _fit_resample(self, X, y):
        random_state = check_random_state(self.random_state)
        X_res = [X.copy()]
        y_res = [y.copy()]

        n_features = X.shape[1]
        prowras_samples = defaultdict(lambda: [])

        if not isinstance(self.std, Number) and len(self.std) != n_features:
            raise ValueError(
                "``std`` and number of features of `X` must be equal."
            )

        self.cdata_ = []
        for minority_class, samples_to_make in self.sampling_strategy_.items():
            if samples_to_make == 0:
                continue

            mask = (y == minority_class)
            cdata = self._make_cluster_data(
                MinorityData(X[mask], [], [], []),
                X[~mask],
                samples_to_make
            )
            self.cdata_.append(cdata)

            cnn = check_neighbors_object(
                "n_cluster_neighbors", self.n_cluster_neighbors
            )

            for i, c in enumerate(cdata.clusters):
                n_samples = cdata.n_new_samples[i]
                if cnn.n_neighbors < c.shape[0]:
                    cnn.set_params(n_jobs=self.n_jobs).fit(c)
                    neighborhoods = c[cnn.kneighbors(c, return_distance=False)]
                    rand_idx = random_state.integers(
                        0, neighborhoods.shape[0], size=n_samples
                    )
                else:
                    neighborhoods = c[None, :, :]
                    rand_idx = [0] * n_samples
                n_affine = (
                    2 if cdata.n_affine[i] < n_features else cdata.n_affine[i]
                )

                if cdata.n_affine[i] < n_features:
                    shadows = neighborhoods[rand_idx].reshape(-1, n_features)
                else:
                    groups = neighborhoods[rand_idx]
                    group_size = groups.shape[1]
                    size = (n_samples, self.n_shadow, group_size, n_features)
                    norms = random_state.normal(scale=self.std, size=size)
                    shadows = groups[:, None, :, :] + norms
                    shadows = shadows.reshape(-1, n_features)

                shadow_idx = random_state.integers(
                    0, shadows.shape[0], size=(n_samples, n_affine)
                )

                affine_weights = random_state.dirichlet(
                    [1] * n_affine, size=n_samples
                )[:, None]
                synthetic_samples = (
                    affine_weights @ shadows[shadow_idx]
                ).reshape(-1, n_features)

                prowras_samples[minority_class].append(synthetic_samples)

            samples_to_drop = sum(cdata.n_new_samples) - samples_to_make
            random_state.shuffle(prowras_samples[minority_class])
            X_res.append(
                np.vstack(prowras_samples[minority_class])[samples_to_drop:]
            )
            y_res.append([minority_class] * X_res[-1].shape[0])

        return np.concatenate(X_res), np.concatenate(y_res)



import numpy as np


def check_random_state(random_state):
    if isinstance(random_state, np.random.RandomState):
        return np.random.default_rng(random_state._bit_generator)
    return np.random.default_rng(random_state)


def safe_random_state(random_state):
    """
    To be used with sklearn Estimators that expect a random_state argument.

    This safely passes the required ``np.random.RandomState`` instance to the
    estimator when the user has a ``np.random.Generator`` instead.
    """
    if isinstance(random_state, np.random.Generator):
        return np.random.RandomState(random_state.bit_generator)
    return random_state



"""
Copyright (c) 2021, Zolisa Bleki
SPDX-License-Identifier: BSD-3-Clause
"""
from collections import defaultdict
from math import ceil

from imblearn.over_sampling.base import BaseOverSampler
from imblearn.utils import Substitution
from imblearn.utils._docstring import (
    _random_state_docstring,
    _n_jobs_docstring,
)
from imblearn.utils._validation import check_neighbors_object
import numpy as np
from sklearn.manifold import TSNE
from sklearn.base import clone

from ._common import check_random_state, safe_random_state



@Substitution(
    sampling_strategy=BaseOverSampler._sampling_strategy_docstring,
    random_state=_random_state_docstring,
    n_jobs=_n_jobs_docstring
)
class LORAS(BaseOverSampler):
    """Localized Random Affine Shadowsampling (LoRAS).

    This class implements the LoRAS oversampling technique for imbalanced
    datasets. This technique generates Gaussian noise in small neighborhoods
    around the minority class samples and then the finaly synthetic samples
    are obtained by a convex combination of multiple noisy data points
    (shadowsamples).

    Parameters
    ----------
    {sampling_strategy}
    n_neighbors : int or estimator object, default=None
        If ``int``, number of nearest neighbours to used to construct synthetic
        samples. If object, an estimator that inherits from
        :class:`~sklearn.neighbors.base.KNeighborsMixin` that will be used to
        find the k_neighbors.
    n_shadow : int, default=None
        The number of shadow samples to generate per minority class data point.
    std : float or sequence, default=0.005
        The standard deviation of the Normal distribution to add to each
        feature when generating shadow samples. If the input is a sequence, its
        size must be equal to the number of features of ``X`` when calling
        the ``fit_resample`` method. If ``float``, then same standard deviation
        will be used for all shadow samples generated.
    n_affine : int, default=None
        The number of shadow samples to use when generating the synthetic
        samples through random affine combinations. If given, the value must be
        between ``2`` and the number of features used in the fitting data.
        If not given, the value will be set to the total number of features in
        fitting data.
    manifold_learner : object, default=None
        An instance of an object that to perform a 2-dimensional embedding of
        a dataset. It must implement the scikit-learn Estimator interface,
        ``fit_transform`` and ``set_params`` methods must be implemented.
        If not given, the :class:`~sklearn.manifold.TSNE` class is used to
        obtain the 2d manifold of the data. Defaults to None.
    manifold_learner_params : dict, default=None
        A dictionary of additional parameters to pass to the instance of the
        ``manifold_learner`` (or TSNE if ``manifold_learner`` is None) when
        creating a 2D manifold of the fitting data. The keys are the parameter
        names and the values are the values. If not given, the default values
        are used.
    {random_state}
    {n_jobs}

    References
    ----------
    .. [1] Bej, S., Davtyan, N., Wolfien, M. et al. LoRAS: an oversampling
       approach for imbalanced datasets. Mach Learn 110, 279–301 (2021).
       https://doi.org/10.1007/s10994-020-05913-4

    Examples
    --------
    >>> from pyloras import LORAS
    >>> from sklearn.datasets import make_classification
    >>> from collections import Counter
    >>> l = LORAS()
    >>> X, y = make_classification(n_classes=3, class_sep=3,
    ... weights=[0.7, 0.2, 0.1], n_informative=3, n_redundant=1, flip_y=0,
    ... n_features=20, n_clusters_per_class=1, n_samples=2000, random_state=10)
    >>> print('Original dataset shape %s' % Counter(y))
    Original dataset shape Counter({{1: 400, 2: 200, 0: 1400}})
    >>> X_res, y_res = l.fit_resample(X, y)
    >>> print(f"Resampled dataset shape % Counter(y_res))
    Resampled dataset shape Counter({{1: 1400, 2: 1400, 0: 1400}})

    """
    def __init__(
        self,
        *,
        sampling_strategy="auto",
        n_neighbors=None,
        n_shadow=None,
        std=0.005,
        n_affine=None,
        manifold_learner=None,
        manifold_learner_params=None,
        random_state=None,
        n_jobs=None
    ):
        super().__init__(sampling_strategy=sampling_strategy)
        self.n_neighbors = n_neighbors
        self.n_shadow = n_shadow
        self.std = std
        self.n_affine = n_affine
        self.manifold_learner = manifold_learner
        self.manifold_learner_params = manifold_learner_params
        self.random_state = random_state
        self.n_jobs = n_jobs

    def _check_2d_manifold_learner(self):
        if (not hasattr(self.manifold_learner, "fit_transform") or
                not hasattr(self.manifold_learner, "set_params")):
            raise ValueError(
                "The 2d manifold learner must implement the ``fit_transform`` "
                "and ``set_params`` methods"
            )

        return clone(self.manifold_learner)

    def _initialize_params(self, X, y, rng):
        """Initialize the parameter values to their appropriate values."""
        f_size = X.shape[1]
        self.n_affine_ = f_size if self.n_affine is None else self.n_affine

        if self.manifold_learner:
            self.manifold_learner_ = self._check_2d_manifold_learner()
        else:
            self.manifold_learner_ = TSNE(n_components=2, perplexity=5)
        if self.manifold_learner_params is not None:
            self.manifold_learner_.set_params(**self.manifold_learner_params)
        try:
            self.manifold_learner_.set_params(random_state=safe_random_state(rng))
        except ValueError:
            pass

        _, y_counts = np.unique(y, return_counts=True)
        if self.n_neighbors is None:
            n_neighbors = 30 if y_counts.min() >= 100 else 5
        else:
            n_neighbors = self.n_neighbors
        self.nn_ = check_neighbors_object("n_neighbors", n_neighbors)
        if self.n_jobs is not None:
            self.nn_.set_params(n_jobs=self.n_jobs)

        if self.n_shadow is None:
            self.n_shadow_ = max(ceil(2 * f_size / self.nn_.n_neighbors), 40)
        else:
            self.n_shadow_ = self.n_shadow

        if self.n_affine_ >= self.nn_.n_neighbors * self.n_shadow_:
            raise ValueError(
                "The number of shadow samples used to create an affine random "
                "combination must be less than `n_neighbors * n_shadow`."
            )

        try:
            iter(self.std)
            self.std_ = self.std
        except TypeError:
            self.std_ = [self.std] * f_size

    def _fit_resample(self, X, y):
        random_state = check_random_state(self.random_state)
        self._initialize_params(X, y, random_state)
        n_features = X.shape[1]

        X_res = [X.copy()]
        y_res = [y.copy()]
        dirichlet_param = [1] * self.n_affine_
        loras_samples = defaultdict(lambda : [])

        for minority_class, samples_to_make in self.sampling_strategy_.items():
            if samples_to_make == 0:
                continue
            X_minority = X[y == minority_class]
            X_embedded = self.manifold_learner_.fit_transform(X_minority)
            self.nn_.fit(X_embedded)
            neighborhoods = self.nn_.kneighbors(X_embedded, return_distance=False)
            num_loras = ceil(samples_to_make / X_embedded.shape[0])
            for neighbor_group in neighborhoods:
                shadow_sample_size = (self.n_shadow_, self.nn_.n_neighbors, n_features)
                total_shadow_samples = (
                    X_minority[neighbor_group] +
                    random_state.normal(scale=self.std_, size=shadow_sample_size)
                ).reshape(self.n_shadow_ * self.nn_.n_neighbors, n_features)
                random_index = random_state.integers(
                    0, total_shadow_samples.shape[0], size=(num_loras, self.n_affine_)
                )
                weights = random_state.dirichlet(dirichlet_param, size=num_loras)
                loras_samples[minority_class].append(
                    (weights[:, None] @ total_shadow_samples[random_index])
                    .reshape(num_loras, n_features)
                )
            # keep only ``samples_to_make`` synthetic samples from the generated.
            samples_to_drop = X_embedded.shape[0] * num_loras - samples_to_make
            random_state.shuffle(loras_samples[minority_class])
            X_res.append(
                np.concatenate(loras_samples[minority_class])[samples_to_drop:]
            )
            y_res.append([minority_class] * samples_to_make)

        return np.concatenate(X_res), np.concatenate(y_res)
